#################################
#    CMFSubversionBrowser
#    Copyright (C) 2002-2004  kapil thangavelu <k_vertigo@objectrealms.net>
#
#    This product is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This product is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this product; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#################################

"""
Purpose: CMF variant of SVN Browser object model
Author: kapil thangavelu <pluto@objectrealms.net>
$Id: CMFSubversionBrowser.py,v 1.3 2003/02/28 05:10:01 k_vertigo Exp $
"""

import os

# zope infrastructure
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem

# cmf infrastructure
from Products.CMFCore.PortalContent import PortalContent
from Products.CMFCore.DynamicType import DynamicType 
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import _getViewFor
from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
from Products.CMFDefault.SkinnedFolder import SkinnedFolder
from Products.CMFCore.interfaces.Dynamic import DynamicType as IDynamicType

from SubversionBrowser import SubversionContext, SubversionDirectory, SubversionFile
from utils import svn_path_id

from Script import RevisionMapGenerator
from Traversal import SubversionZopeTraversal
from TypeInfo import svn_file_fti, svn_directory_fti

class PortalView:
    
    index_html = None
    
    def __call__(self):
        '''
        Invokes the default view.
        '''
        view = _getViewFor(self)
        if getattr(aq_base(view), 'isDocTemp', 0):
            return apply(view, (self, self.REQUEST))
        else:
            return view()

    def view(self):
        '''
        invoke default view even if index_html is overwritten.
        '''
        return self()
    
class CMFSubversionDirectory(SubversionDirectory,
                             PortalView,
                             DynamicType,
                             SimpleItem):

    meta_type = 'CMF Subversion Directory'

    __implements__ = IDynamicType


class CMFSubversionFile(SubversionFile,
                        PortalView,
                        DynamicType,
                        SimpleItem):

    meta_type = 'CMF Subversion File'

    __implements__ = IDynamicType

    def getAnnotated(self, revision_set=None):

        if revision_set is None: # sigh cheesy
            revision_set = ()
        elif isinstance(revision_set, (tuple,list)):
            try: revision_set = map(int, revision_set)
            except: revision_set = ()
        else:
            revision_set = ()
                
        lines = self.getAnnotatedLines(revision_set)
        if not lines:
            return """ Annotation Not Available """

        buf = []
        
        # sigh.. we do a two pass alg here, too identify the biggest number
        # and largest author name so we can do nice margin justification.
        big_rev = 0
        big_auth = 0
        for l in lines:
            if l[0] > big_rev:
                big_rev = l[0]
            if len(l[2]) > big_auth:
                big_auth = len(l[2])

        big_auth += 1
        big_rev = len(str(big_rev))+1

        for l in lines:
            buf.append('%s %s :: %s'%(str(l[0]).ljust(big_rev),
                                      l[2].ljust(big_auth),
                                      l[1]))

        return ''.join(buf)
    
class CMFSubversionBrowser(SubversionContext,
                           SubversionZopeTraversal,
                           SkinnedFolder,
                           PortalContent,
                           DefaultDublinCoreImpl):

    meta_type = 'CMF Subversion Browser'

    __implements__ = PortalContent.__implements__

    security = ClassSecurityInfo()
    security.declareProtected(CMFCorePermissions.ModifyPortalContent, 'editContext')

    security.declarePrivate('SVNDirectoryFactory')
    def SVNDirectoryFactory(self, svn_path):
        id = svn_path_id( svn_path )
        ob = CMFSubversionDirectory( id, svn_path )
        ob._setPortalTypeName(svn_directory_fti['id'])
        return ob

    security.declarePrivate('SVNFileFactory')
    def SVNFileFactory(self, svn_path):
        id = svn_path_id( svn_path )
        ob = CMFSubversionFile( id, svn_path )
        ob._setPortalTypeName(svn_file_fti['id'])
        return ob

    #################################
    # view based support methods
    security.declarePublic('formatLogMessage')
    def formatLogMessage(self, message):
        """
        format a log message such that we only show an abbreviated portion
        as nesc. instead of the entire message, if the message is truncated in
        this manner show elispses.
        """
        msg_len = len(message)
        if msg_len > 60:
            display_message = "%s..."%message[:60]
        else:
            display_message = message
        return display_message

    security.declarePublic('formatRevPath')
    def formatRevPath(self, path, revision=None):
        """
        format an svn path with an optional revision into a url, if the
        url is not below the browser/context then an empty string is returned.
        """

        if not path.startswith(self.svn_path):
            return ''

        url = self.absolute_url()
        remainder_path = path[len(self.svn_path):]
        if not remainder_path.startswith('/'):
            remainder_path = '/%s'%remainder_path

        if revision is not None and isinstance(revision, int):
            revision_part = '?revision:int=%d'%revision
        else:
            revision_part = ''

        url = '%s%s%s'%(url, remainder_path, revision_part)
        return url

    security.declarePublic('formatRevAction')
    def formatRevAction(self, object_url, action, request):
        pass

    security.declarePublic('formatRevisionMap')
    def formatRevisionMap(self):
        """
        display a graphical revision map
        """
        generator = RevisionMapGenerator( self, self.getRevisionInfo(), self.getRevision())
        return generator()
    
InitializeClass(CMFSubversionBrowser)


#################################
# factories for cmf

class ConstructionNotAllowedError(Exception): pass

def null_factory(self, *args, **kw):
    raise ConstructionNotAllowedError("this type cannot be constructed")

def addCMFSubversionBrowser(self, id):
    """ adds svn browsers """
    id = str(id)
    ob = CMFSubversionBrowser(id)
    self._setObject(id, ob)
    
