##################################
#    CMFSubversionBrowser
#    Copyright (C) 2002-2004 kapil thangavelu <k_vertigo@objectrealms.net>
#
#    This product is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This product is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this product; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#################################

"""
Purpose: Initialize and Register the Browser with Zope/CMF
Author: kapil thangavelu <pluto@objectreamls.net>
$Id: SubversionBrowser.py,v 1.3 2003/02/28 05:14:58 k_vertigo Exp $

Details

    - Offer a browser into an svn repository via zope

    - Explore svn apis

this module can be imported outside of zope provided the support modules
are on the PYTHONPATH.
"""

import atexit, os, difflib, time

# our utilities
from utils import svn_path_id, svn_path_join, charbuffer, wrap

# svn bindings
from svn import fs, repos, util

# zope infrastructure
from Acquisition import Implicit
from AccessControl import ClassSecurityInfo, Permissions
from ComputedAttribute import ComputedAttribute
from DateTime import DateTime
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from zLOG import LOG


#################################
# Globals
DEBUG = 0
CHUNK_SIZE = 16384
_marker = object()

#################################
# log class
class _log:

    LOG_NAME = '[SubversionBrowser]'
    LOG_LEVEL = 100

    def debug(self, msg, extra='', **kwargs):
        if not DEBUG:
            return
        #print msg
        LOG(self.LOG_NAME,
            self.LOG_LEVEL,
            msg,
            extra)

log = _log()

#################################
# APR Jaunx

def initializeAPR():
    log.debug('Initializing APR')
    util.apr_initialize()
    def apr_exit(): util.apr_terminate()
    atexit.register(apr_exit)

initializeAPR()

#################################
# Exceptions

class InvalidRepositoryPath(Exception): pass
class RootPathViolation(Exception): pass
class UnsupportedNodeType(Exception): pass

#################################
# SVN Resource Bag

class ResourceContext:
    
    """
    All Subversion/APR resources held/used by an svn context are stored in
    a resource context. The encapsulation in a resource context allows for
    tranparent cleanup of variables for finalization methods when the volatile
    variable reference used by the svn context goes out of scope or is garbage
    collected.
    """
    
    def __init__(self, repository_path, svn_path, revision=None):
        self.repository_path = repository_path
        self.svn_path = repository_path

        self.finalized = False
        self.initialized = False

        self.pool = None
        self.repository = None
        self.fsptr = None
        self.fsroot = None
        self.revision = revision

        self.initialize() 

    def setRevision(self, revision=None):
        sr = revision
        if revision is not None:
            revision = int(revision)
        else:
            revision = fs.youngest_rev( self.fsptr, self.pool)
        if self.fsroot is not None:
            fs.close_root( self.fsroot )

        self.fsroot = fs.revision_root( self.fsptr,
                                        revision,
                                        self.pool )
        self.revision = revision
        
    def initialize(self):
        if self.initialized:
            return

        if self.revision is not None:
            if not isinstance(self.revision, int):
                try:
                    self.revision = int(self.revision)
                except:
                    raise SyntaxError("Revisions must be specified as integers")
            
        log.debug('initializing')

        self.pool = util.svn_pool_create( None )
        self.repository = repos.svn_repos_open(self.repository_path, self.pool)
        self.fsptr = repos.svn_repos_fs( self.repository )

        if self.revision is None:
            self.revision = fs.youngest_rev( self.fsptr, self.pool)            

        self.setRevision( self.revision )
        self.initialized = True

    def finalize(self):
        if self.finalized or not self.initialized:
            return
        fs.close_root( self.fsroot )
        util.svn_pool_clear( self.pool )
        util.svn_pool_destroy( self.pool )
        self.finalized = True
        
    def __del__(self):
        self.finalize()

#################################
# svn browser

class SubversionContext(SimpleItem):
    """
    a window into a subversion repository.
    """
    
    meta_type = 'Subversion Context'

    security = ClassSecurityInfo()

    _v_resource_context = None

    def __init__(self, id, repository_path='', svn_path=''):
        self.id = id
        self.repository_path = repository_path
        self.svn_path = svn_path  # the path we should start browsing at
        
    security.declarePrivate('initialize')
    def initialize(self, revision=None):
        """
        initialize the browser, with a new pool, revison root, and repository
        pointer. optional revision arg can specify the revision root
        """

        if self._v_resource_context and \
           ((revision is None) or \
             revision==self._v_resource_context.revision):
            return self._v_resource_context

        self._v_resource_context = rc = ResourceContext( self.repository_path,
                                                         self.svn_path,
                                                         revision )
        return rc

    security.declarePrivate('getResourceContext')
    def getResourceContext(self):
        return self._v_resource_context
    
    security.declarePrivate('setRevision')
    def setRevision(self, revision=None):
        """
        change the revision root of the fs being inspected
        
        # setrevision is used as a callback from traversal, so it performs
        # a sanity check ones state, as a newly created svn context does not
        # have adequate state to be access a repository.
        """
        if not self.repository_path:
            return


        if not self._v_resource_context:
            return self.initialize( revision )
            
        #print 'Set Revision', repr(revision), self._v_resource_context.revision
        
        self._v_resource_context.setRevision( revision )

    security.declarePublic('getRevision')
    def getRevision(self):
        """
        return the revision currently being introspected.
        """
        resource_context = self.initialize()
        return resource_context.revision

    security.declareProtected(Permissions.view, 'getRevisionInfo')
    def getRevisionInfo(self):
        """
        return a plethora of info about when and what was committed in this
        revision.
        """
        svn_ctx = self.getSVNContext()
        resource_ctx = svn_ctx.initialize()
        revision = self.getRevision()
        info = SVNLogFactory(resource_ctx.fsptr, revision, resource_ctx.pool)
        info.paths = fs.paths_changed( resource_ctx.fsroot, resource_ctx.pool )
        return info

    security.declarePrivate('clear')
    def clear(self):
        """
        reset any associated svn/apr resources
        """
        if self._v_resource_context is not None:
            self._v_resource_context.finalize()
            del self._v_resource_context
        
    security.declarePrivate('getSVNObject')
    def getSVNObject(self, path):
        """
        retrieve a subversion object from an svn path. does
        not wrap the object.
        """
        log.debug('retrieving %s'%str(path))

        resource_context = self.initialize()
        
        if fs.is_dir(resource_context.fsroot, path, resource_context.pool):
            return self.SVNDirectoryFactory(path)
        elif fs.is_file(resource_context.fsroot, path, resource_context.pool):
            return self.SVNFileFactory(path)
        else:
            return None

    security.declarePublic('getSVNContext')
    def getSVNContext(self):
        return self

    security.declarePrivate('SVNDirectoryFactory')
    def SVNDirectoryFactory(self, svn_path):
        """
        contruct an svn directory node from the given svnpath
        """
        id = svn_path_id( svn_path )
        return SubversionDirectory( id, svn_path )

    security.declarePrivate('SVNFileFactory')
    def SVNFileFactory(self, svn_path):
        """
        contruct an svn file node from the given svnpath
        """
        id = svn_path_id( svn_path )
        return SubversionFile( id, svn_path )

    #################################
    # public api methods

    security.declareProtected(Permissions.view, 'getRootSVNObject')
    def getSVNRootObject(self):
        """
        get the root object that the svn context/window is anchored to
        """
        return self.getSVNObject(self.svn_path).__of__(self)

    def editContext(self, repository_path, svn_path):
        """
        edit the settings for the subversion context.
        """
        if not os.path.exists(repository_path):
            raise InvalidRepositoryPath(repository_path)
        
        self.repository_path = repository_path
        self.svn_path = svn_path

        # try and open the repository just to be sure
        self.initialize()

InitializeClass(SubversionContext)

#################################
# svn object model        

class SubversionObject(Implicit):

    security = ClassSecurityInfo()

    def _v_prop_cache(self):
        self._v_prop_cache = cache = {}
        return cache

    _v_prop_cache = ComputedAttribute( _v_prop_cache )

    
    def __init__(self, id=None, svn_path=None):
        self.id = id
        self.svn_path = svn_path

    security.declarePublic('getId')
    def getId(self):
        return self.id

    security.declarePrivate('getNodeId')
    def getNodeId(self):
        """
        Return the Subversion Node Id
        """
        svn_ctx = self.getSVNContext()
        resource_ctx = svn_ctx.initialize()
        uid = fs.node_id(resource_ctx.fsroot, self.svn_path, resource_ctx.pool)
        return '%s'%fs.unparse_id(uid, resource_ctx.pool)

    security.declareProtected(Permissions.view, 'getLog')
    def getLog(self, revision=None):
        """ 
        Return A Log Entry Object for A Single Revision. If Revision
        is not specified then the revision is the latest available.
        """
        svn_ctx = self.getSVNContext()
        resource_ctx = svn_ctx.initialize()
        
        if revision is None:
            revision = fs.node_created_rev(resource_ctx.fsroot,
                                           self.svn_path,
                                           resource_ctx.pool)
        
        return SVNLogFactory(resource_ctx.fsptr, revision, resource_ctx.pool)

    security.declareProtected(Permissions.view, 'getLogEntries')
    def getLogEntries(self, changed_paths=0):
        """
        get all log entries associated with this node
        """
        
        svn_ctx = self.getSVNContext()
        resource_ctx = svn_ctx.initialize()
        collection = SVNLogCollection()
        
        repos.svn_repos_get_logs(
            resource_ctx.repository,  # repository
            [self.svn_path],           # paths
            resource_ctx.revision,    # start revision
            0,                        # end revision (for sort)
            changed_paths,            # discover changed paths 
            0,                        # strict node history (if true)
            collection.addLog,        # reciever
            resource_ctx.pool)

        return collection.getLogs()

    def getMappedLogEntries(self):
        """
        get all log entries associated with this node, and get the path
        this node was at for each revision log.
        """
        entries = self.getLogEntries()
        rev_map = self.getRevisionPathMap(cross_copies=1)
        for e in entries:
            e.rev_path = rev_map[e.revision]
        return entries

    def getParent(self):
        """
        get the parent object of this node
        XXX should we should just be using acq contexts
        """
        svn_ctx = self.getSVNContext()
        segments = self.svn_path.split('/')
        segments.pop(-1)
        svn_parent_path = svn_path_join(*segments)
        if not len(svn_ctx.svn_path) >= len(svn_parent_path):
            raise RootPathViolation("SVN Parent Path Out of Scope")
        return svn_ctx.getSVNObject(svn_parent_path).__of__(svn_ctx)

    security.declarePublic('getProperty')
    def getProperty(self, property_name): 
        """
        retrieve a given property name
        """
        svn_ctx = self.getSVNContext()
        revision = svn_ctx.getRevision()
         
        cache = self._v_prop_cache.setdefault(revision, {})
        value = cache.get(property_name, _marker)
        if value is not _marker:
            return value

        resource_ctx = svn_ctx.initialize()
        
        value = fs.node_prop(resource_ctx.fsroot, # root
                             self.svn_path,       # path 
                             property_name,       # prop name
                             resource_ctx.pool    # pool
                             )
        cache[property_name] = value
        return value

    security.declarePublic('getProperties')
    def getProperties(self):
        """
        return all the node property names
        """
        svn_ctx = self.getSVNContext()
        revision = svn_ctx.getRevision()

##         XXX hmm.. caching here creates ordering deps with the above
##         ie getting a specific property will mean only that prop gets found
##         in a discovery of all props. urgh. sigh.. i'm willing to live with
##         it for now though.

##         cache = self._v_prop_cache.setdefault(revision, {})
##         if cache is not None:
##             return cache.keys()
        
        resource_ctx = svn_ctx.initialize()
        prop_names =  fs.node_proplist(resource_ctx.fsroot,
                                       self.svn_path,
                                       resource_ctx.pool)
##         for pn in prop_names:
##             cache[pn] = _marker
        return prop_names

    def getRevisionPathMap(self, cross_copies=0):
        """
        for every revision this node was modified return a dictionary
        mapping each revision to the path the node was at.
        """
        svn_context = self.getSVNContext()
        resource_ctx = svn_context.initialize()
        pool = util.svn_pool_create(resource_ctx.pool)
        rev_map = {}
        history = fs.node_history( resource_ctx.fsroot, self.svn_path, pool)

        try:
            while history:
                history = fs.history_prev( history, cross_copies, pool )
                if history is None: break
                path, rev = fs.history_location( history, pool )
                rev_map[ rev ] = path
        finally:
            util.svn_pool_destroy( pool )
            
        return rev_map
        
    def getRevisionCreated(self):
        """
        return the revision that this node was created.
        """
        svn_ctx = self.getSVNContext()
        resource_ctx = svn_ctx.initialize()

        pool = util.svn_pool_create(resource_ctx.pool)
        try:
            revision = fs.node_created_rev(resource_ctx.fsroot,
                                           self.svn_path,
                                           pool)
        finally:
            util.svn_pool_destroy(pool)

        return revision
    
    security.declarePublic('isBinary')
    def isBinary(self):
        """
        does the node represent binary content 
        """
        mime_type = self.getProperty(util.SVN_PROP_MIME_TYPE)
        if mime_type is None:
            return 0        
        return util.svn_mime_type_is_binary(mime_type)
        
    security.declarePublic('absolute_url')
    def absolute_url(self, relative=0):
        
        svn_context = self.getSVNContext()
        url = svn_context.absolute_url(relative=relative)
        relative_path = self.svn_path[len(svn_context.svn_path)+1:]
        
        return "%s/%s"%(url, relative_path)
        

InitializeClass(SubversionObject)

class SubversionDirectory(SubversionObject):
    """
    Subversion Directory Node
    """
    
    meta_type = 'Subversion Directory'

    security = ClassSecurityInfo()

    isSVNDirectory = 1

    def _v_dirent_cache(self):
        self._v_dirent_cache = cache = {}
        return cache

    _v_dirent_cache = ComputedAttribute( _v_dirent_cache )

    def __getitem__(self, key):
        """
        mapping interface for subnode retrieval
        """
        svn_context = self.getSVNContext()
        resource_ctx = svn_context.initialize()
        path = svn_path_join(self.svn_path, key)

        try:
            if fs.is_dir(resource_ctx.fsroot, path, resource_ctx.pool):
                return svn_context.SVNDirectoryFactory(path).__of__(self)
            if fs.is_file(resource_ctx.fsroot, path, resource_ctx.pool):
                return svn_context.SVNFileFactory(path).__of__(self)
        except:
            return getattr(self, key)

    security.declarePrivate('getNodes')
    def getNodes(self, node_type=None):
        """
        node_type is the svn node type we should return results for, if not
        specified return all nodes. automatically caches results.
        """
        svn_context = self.getSVNContext()
        revision = svn_context.getRevision()

        if node_type is not None and (not isinstance(node_type, int) or node_type > 3 or node_type < 0):
            raise UnsupportedNodeType("%r"%node_type)

        entries =  self._v_dirent_cache.get( revision )

        if entries is None:
            self._fillNodeCache()
            entries = self._v_dirent_cache.get( revision )
            
        if node_type is None:
            res = []
            res.extend( entries[util.svn_node_dir] )
            res.extend( entries[util.svn_node_file] )
            return res
        elif node_type == util.svn_node_dir:
            return entries[util.svn_node_dir]
        elif node_type == util.svn_node_file:
            return entries[util.svn_node_file]

    def _fillNodeCache(self):

        svn_context = self.getSVNContext()
        resource_ctx = svn_context.initialize()
        revision = svn_context.getRevision()
        
        entries = fs.dir_entries( resource_ctx.fsroot,
                                  self.svn_path,
                                  resource_ctx.pool)

        cache = [ (), [], [] ]
        names = entries.keys()
        names.sort()
        
        for name in names:
            entry = entries[name]
            entry_path = svn_path_join( self.svn_path, name )
            if entry.kind == util.svn_node_dir:
                cache[util.svn_node_dir].append(
                    svn_context.SVNDirectoryFactory( entry_path )
                    )
            elif entry.kind == util.svn_node_file:
                cache[util.svn_node_file].append(
                    svn_context.SVNFileFactory( entry_path )
                    )
        self._v_dirent_cache[ revision ] = cache

    security.declareProtected(Permissions.view, 'getDirectories')
    def getDirectories(self):
        """
        get all the subdirectories of this node
        """
        directories = self.getNodes( util.svn_node_dir )
        wrapper = wrap( self )
        return [wrapper( dir_node ) for dir_node in directories ]
        
    security.declareProtected(Permissions.view, 'getFiles')
    def getFiles(self):
        """
        get all the files in this directory
        """
        files = self.getNodes( util.svn_node_file )
        wrapper = wrap( self )
        return [wrapper( file_node ) for file_node in files ]

InitializeClass(SubversionDirectory)

class SubversionFile(SubversionObject):
    """
    Subversion File Node
    """
    
    meta_type = 'Subversion File'
    
    security = ClassSecurityInfo()

    isSVNFile = 1

    security.declarePublic('getMimeType')
    def getMimeType(self):
        """
        get the mime type of this file node if set, if not return
        text/plain
        """
        mime_type = self.getProperty(util.SVN_PROP_MIME_TYPE)
        if mime_type is None:
            return 'text/plain'        
        return mime_type
    
    security.declareProtected(Permissions.view, 'getContents')
    def getContents(self, revision=None, writer=None):
        """
        get the contents of the file node, if writer is provide, it should
        be a file handle like interface to which the contents of the file
        will be streamed, else the entire file contents will be returned
        as a string.
        """
        rval = not writer
        svn_context = self.getSVNContext()

        if revision is not None and isinstance(revision, int):
            resource_ctx = svn_context.initialize(revision)
        else:
            resource_ctx = svn_context.initialize()
        
        stream = fs.file_contents( resource_ctx.fsroot,
                                   self.svn_path,
                                   resource_ctx.pool )

        if writer is None:
            writer = charbuffer()

        while 1:
            data = util.svn_stream_read(stream, CHUNK_SIZE)
            if not data:
                break
            writer.write(data)

        if revision is not None:
            svn_context.clear()

        if rval:
            return writer.getvalue()

    security.declareProtected(Permissions.view, 'getAnnotatedLines')
    def getAnnotatedLines(self, revision_set=(), include_copies=1):
        """
        perform a diff of the file node contents, returns an array
        of ( line number, (revision, line content, author ) ).
        
        revision set if passed in, should be revisions that the content
        was changed in, only these revisions will be considered, else
        all revisions the content was altered in will be considered.

        if included copies is true than cross copy history when finding
        revisions to include in diff.
        """
        stime = time.time()
        svn_context = self.getSVNContext()
        annotations = []
        contents = ''

        revision_path_map = self.getRevisionPathMap( include_copies )

        if not revision_set:
            revision_set = revision_path_map.keys()

        # set revision order low high
        revision_set.sort()
        # flush mem.
        resource_ctx = svn_context.initialize(revision_set[-1])

        if not len(revision_set) > 1 or self.isBinary():
            return ()

        pool = util.svn_pool_create(resource_ctx.pool)
        
        try:
            for r in revision_set:
                revision_root = fs.revision_root(resource_ctx.fsptr, r, pool)
                
                rev_author = fs.revision_prop(resource_ctx.fsptr,
                                              r,
                                              util.SVN_PROP_REVISION_AUTHOR,
                                              pool)
                
                stream = fs.file_contents(revision_root,
                                          revision_path_map[r],
                                          pool)
                buffer = charbuffer()

                while 1:
                    data = util.svn_stream_read(stream, CHUNK_SIZE)
                    if not data:
                        break
                    buffer.write(data)

                previous_contents = contents
                contents = buffer.getvalue()
                delta = difflib.ndiff(
                    previous_contents.splitlines(1),
                    contents.splitlines(1)
                    )

                # annotate algorithm
                li = 0
                for d in delta:
                    flag = d[0]
                    if flag == ' ':
                        li += 1
                        continue
                    elif flag == '?':
                        continue
                    elif flag == '+':
                        annotations.insert(li, (r, d[2:], rev_author) )
                        li += 1
                    elif flag == '-':
                        del annotations[li]

                fs.close_root(revision_root)
        finally:
            util.svn_pool_clear(pool)
            util.svn_pool_destroy(pool)

        # reset the context
        svn_context.clear()
        return [annotations[i] for i in xrange(len(annotations))]
            

InitializeClass(SubversionFile)        
        
class SubversionLogEntry:

    __allow_access_to_unprotected_subobjects__ = 1
    
    def __init__(self, **args):
        self.__dict__.update(args)

    def __repr__(self):
        return "<Log: %s>"%' '.join( self.__dict__.keys() )

#################################
# factories        

def SVNLogFactory(fsptr, revision, pool):
    
    author  = fs.revision_prop(fsptr, revision, util.SVN_PROP_REVISION_AUTHOR, pool)
    message = fs.revision_prop(fsptr, revision, util.SVN_PROP_REVISION_LOG, pool)
    date    = fs.revision_prop(fsptr, revision, util.SVN_PROP_REVISION_DATE, pool)
    
    date = util.svn_time_from_cstring(date, pool)

    return SubversionLogEntry(author = author,
                              message = message,
                              date = zope_time(date),
                              revision = revision
                              )

class SVNLogCollection:

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self):
        self.logs = []

    def addLog(self, paths, revision, author, date, message, pool):
        """
        Used for log gather callbacks
        
        notes in svn_repos.h, svn_types.h, svn_clients.h 
        
        typedef svn_error_t *(*svn_log_message_receiver_t)
        (void *baton,
        apr_hash_t *changed_paths,
        svn_revnum_t revision,
        const char *author,
        const char *date,  /* use svn_time_from_string() if need apr_time_t */
        const char *message,
        apr_pool_t *pool);
        """
        
        date = util.svn_time_from_cstring(date, pool)
    
        self.logs.append( SubversionLogEntry(author = author,
                                             message = message,
                                             date = zope_time(date),
                                             revision = revision,
                                             paths = paths
                                             )
                          )

    def __getitem__(self, i):
        return self.logs[i]
    

    def getLogs(self):
        return self.logs

def zope_time(aprtime):
    return DateTime(aprtime/1000000)

